﻿module CGameController;
import std.algorithm, std.conv, std.math, std.random;
import DOpenGL.gl;
import Obelisk.CColour, Obelisk.CVector, Obelisk.EKey, Obelisk.IObelisk;
import CPlanet, CShip, CView;

class CGameController
{
	private IObelisk mObelisk;
	private CView mView;

	private const float cSpaceSize = 20000;

	private CShip mShip;
	private CPlanet[] mPlanets;

	private CPlanet mNearPlanet = null;
	private bool mCanLand;
	private bool mDontLand = false;

	private int mFuel = 100;
	private int mFuelUsed = 0;
	private int mAmmo = 20;
	private int[3] mCargo = [ 0, 0, 0 ];

	private CPlanet mShowUI = null;
	private int mTradeResource = -1;

	private string mInfoText = "";
	private bool mGameOver = false;

	private int mThrustCount = -1;

	public this(IObelisk aObelisk)
	{
		mObelisk = aObelisk;
		mView = new CView(mObelisk, cSpaceSize);

		mObelisk.SetBackgroundColour(0.06, 0.06, 0.12);
		GenerateStars();
		GeneratePlanets();
		mPlanets[0].GetViewObject().MinimapColour = new CColour(0x3D / 255.0f, 0xE5 / 255.0f, 0x46 / 255.0f);
		mShip = new CShip(mView, mPlanets[0].GetX(), mPlanets[0].GetY() - 96 - 16);
		mShip.GetViewObject().MinimapColour = new CColour(0.6956, 0.9592, 0.7098);
		mView.SetCameraObject(mShip.GetViewObject());
		mShowUI = mPlanets[0];
	}



	private void Thrust()
	{
		mShip.Thrust();
		BurnFuel();
		mView.ShowThrust(true);
	}



	private void ThrustOn()
	{
		mThrustCount = 18;
	}



	private void BurnFuel()
	{
		mFuelUsed += 4;
		if (mFuelUsed >= 100)
		{
			if (--mFuel == 0)
			{
				if (mPlanets[0].GetResource(0) < 10)
				{
					mGameOver = true;
					mInfoText = "Game Over";
				}
				else
				{
					mInfoText = "You ran out of fuel.";
				}
			}
			mFuelUsed -= 100;
		}
	}



	private void Refuel()
	{
		int FuelAmount = min(mPlanets[0].GetResource(0), 100 - mFuel);
		mFuel += FuelAmount;
		mPlanets[0].AddResource(0, -FuelAmount);
	}



	public void Update()
	{
		if (mInfoText != "")
		{
			if (mGameOver || (!mObelisk.IsKeyDown(EKey.Enter) && !mObelisk.MouseButtonClicked(false)))
				return;
			mInfoText = "";
			mPlanets[0].AddResource(2, -25);
			mShip.Reset(mPlanets[0].GetX(), mPlanets[0].GetY() - 96 - 16);
			for (int i = 0; i < 3; i++)
				mCargo[i] = 0;
			mShowUI = mPlanets[0];
			Refuel();
		}

		mView.ShowThrust(false);

		if (mShowUI !is null)
		{
			int GuiX = 512 - 650 / 2;
			int GuiY = 384 - 450 / 2;
			if (ButtonClicked(GuiX + 650 - 50 - 150, GuiY + 95, 150, 30))
			{
				if (mShowUI is mPlanets[0])
				{
					int Amount = min(10, mShowUI.GetResource(0));
					mShowUI.AddResource(0, -Amount);
					mCargo[0] += Amount;
				}
				else
					mTradeResource = 0;
			}
			else if (ButtonClicked(GuiX + 650 - 50 - 150, GuiY + 95 + 35, 150, 30))
			{
				if (mShowUI is mPlanets[0])
				{
					int Amount = min(5, mShowUI.GetResource(1));
					mShowUI.AddResource(1, -Amount);
					mCargo[1] += Amount;
				}
				else
					mTradeResource = 1;
			}
			else if (ButtonClicked(GuiX + 650 - 50 - 150, GuiY + 95 + 70, 150, 30))
			{
				if (mShowUI is mPlanets[0])
				{
					int Amount = min(10, mShowUI.GetResource(2));
					mShowUI.AddResource(2, -Amount);
					mCargo[2] += Amount;
				}
				else
					mTradeResource = 2;
			}
			else if (ButtonClicked(512 - 250 / 2, GuiY + 450 - 20 - 45, 250, 45))
			{
				mShowUI = null;
				ThrustOn();
				mDontLand = true;
				mTradeResource = -1;
			}
			if (mTradeResource != -1)
			{
				if (mTradeResource != 0 && mCargo[0] != 0 &&
						ButtonClicked(GuiX + 650 - 50 - 150, GuiY + 95 + 170, 150, 30))
					Trade(0, 10);
				else if (mTradeResource != 1 && mCargo[1] != 0 &&
						ButtonClicked(GuiX + 650 - 50 - 150, GuiY + 95 + 205, 150, 30))
					Trade(1, 5);
				else if (mTradeResource != 2 && mCargo[2] != 0 &&
						ButtonClicked(GuiX + 650 - 50 - 150, GuiY + 95 + 240, 150, 30))
					Trade(2, 10);
			}
		}
		else
		{
			if (mThrustCount == -1 && mObelisk.IsKeyDown(EKey.Up))
				Thrust();
			bool Left = mObelisk.IsKeyDown(EKey.Left);
			bool Right = mObelisk.IsKeyDown(EKey.Right);
			if (Left && !Right)
				mShip.Turn(-1);
			else if (Right && !Left)
				mShip.Turn(1);
			mShip.Update();
		}
		if (mThrustCount != -1)
		{
			mThrustCount--;
			Thrust();
		}
		mNearPlanet = null;
		foreach (CPlanet Planet; mPlanets)
		{
			Planet.Update();

			// check for ship/planet collections
			float DeltaX = mView.GetDelta(Planet.GetX(), mShip.GetX());
			float DeltaY = mView.GetDelta(Planet.GetY(), mShip.GetY());
			if (DeltaX ^^ 2 + DeltaY ^^ 2 < 420 ^^ 2)
			{
				if (!mShip.IsLanded() && !mDontLand)
					mShip.Gravity(DeltaX, DeltaY);
				mNearPlanet = Planet;
				float Angle = atan2(DeltaY, DeltaX);
				mCanLand = mShip.GetVelocity() < 25 && AngleBetween(mShip.GetAngle(),
						Angle) < 8 * PI / 180;
				if (DeltaX ^^ 2 + DeltaY ^^ 2 < (96 + 16) ^^ 2)
				{
					if (!mShip.IsLanded() && !mDontLand)
					{
						if (mCanLand)
						{
							// successful landing
							mShip.Land();
							mShowUI = Planet;
							if (Planet is mPlanets[0])
							{
								for (int i = 0; i < 3; i++)
								{
									Planet.AddResource(i, mCargo[i]);
									mCargo[i] = 0;
								}

								Refuel();
							}
						}
						else
						{
							// crash
							if (mPlanets[0].GetResource(2) < 25)
							{
								mGameOver = true;
								mInfoText = "Game Over";
							}
							else
								mInfoText = "You crashed.";
						}
					}
				}
				else
					mDontLand = false;
				break;
			}
		}
	}



	/**
	Sells the specified resource and buys the currently selected trade resource from the
	planet.
	*/
	private void Trade(int aResource, int aMaximum)
	{
		int SellAmount = min(mCargo[aResource], aMaximum);
		int PlanetSellAmount = mShowUI.GetResource(aResource);
		int PlanetBuyAmount = mShowUI.GetResource(mTradeResource);
		int BuyAmount = cast(int)round(SellAmount * GetExchange(aResource, PlanetSellAmount,
				mTradeResource, PlanetBuyAmount));
		if (BuyAmount > PlanetBuyAmount)
		{
			SellAmount /= 2;
			BuyAmount = cast(int)round(SellAmount * GetExchange(aResource, PlanetSellAmount,
					mTradeResource, PlanetBuyAmount));
			if (BuyAmount > PlanetBuyAmount)
			{
				SellAmount /= 2;
				BuyAmount = cast(int)round(SellAmount * GetExchange(aResource, PlanetSellAmount,
						mTradeResource, PlanetBuyAmount));
				if (BuyAmount > PlanetBuyAmount)
					return;
			}
		}
		mCargo[aResource] -= SellAmount;
		mShowUI.AddResource(aResource, SellAmount);
		mCargo[mTradeResource] += BuyAmount;
		mShowUI.AddResource(mTradeResource, -BuyAmount);
	}



	private float AngleBetween(float aAngle0, float aAngle1)
	{
		float Delta = abs(aAngle0 - aAngle1);
		return Delta > PI ? PI * 2 - Delta : Delta;
	}



	public void Render()
	{
		mView.Render();

		// render UI
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(-512, -384, 0);
		mObelisk.SetFont("D3 Euronism_20.png");
		mObelisk.DrawText("Fuel: " ~ to!string(mFuel), 5, 5);
		mObelisk.DrawText("Ammo: " ~ to!string(mAmmo), 5, 40);
		mObelisk.DrawText("Velocity: " ~ to!string(min(cast(int)mShip.GetVelocity(), 99)), 5, 75);
		if (mNearPlanet !is null)
		{
			mObelisk.DrawText("Landing: ", 5, 110);
			DrawRect(215, 115, 40, 25, mCanLand ? 0x53E83C : 0xE12419);
		}
		mObelisk.DrawText("Cargo", 5, 625);
		mObelisk.DrawText("Fuel: " ~ to!string(mCargo[0]), 5, 660);
		mObelisk.DrawText("Ammo: " ~ to!string(mCargo[1]), 5, 695);
		mObelisk.DrawText("Ship Parts: " ~ to!string(mCargo[2]), 5, 730);
		/*
		mObelisk.DrawText("Earth", 630, 625);
		mObelisk.DrawText("Fuel: " ~ to!string(mPlanets[0].GetResource(0)), 630, 660);
		mObelisk.DrawText("Ammo: " ~ to!string(mPlanets[0].GetResource(1)), 630, 695);
		mObelisk.DrawText("Ship Parts: " ~ to!string(mPlanets[0].GetResource(2)), 630, 730);
		*/

		/*
		mObelisk.DrawText("Resource: " ~ to!string(mPlanets[0].GetProduceResource()), 5, 320);
		mObelisk.DrawText("Rate: " ~ to!string(mPlanets[0].GetProduceRate()), 5, 355);
		mObelisk.DrawText("Production: " ~ to!string(mPlanets[0].GetProduction), 5, 390);
		//*/

		if (mShowUI !is null)
		{
			int GuiX = 512 - 650 / 2;
			int GuiY = 384 - 450 / 2;
			DrawRect(GuiX, GuiY, 650, 450, 0xAFAFAF);
			DrawRect(512 - 320 / 2, GuiY + 20, 320, 50, 0x000000);
			mObelisk.DrawText(mShowUI.GetName(), 512 - 320 / 2 + 10, GuiY + 30);

			DrawRect(GuiX + 40, GuiY + 90, 650 - 80, 450 - 90 - 80, 0x000000);
			mObelisk.DrawText("Fuel: " ~ to!string(mShowUI.GetResource(0)), GuiX + 45, GuiY + 95);
			mObelisk.DrawText("Ammo: " ~ to!string(mShowUI.GetResource(1)), GuiX + 45, GuiY + 95 + 35);
			mObelisk.DrawText("Ship Parts: " ~ to!string(mShowUI.GetResource(2)), GuiX + 45, GuiY + 95 + 70);
			string Text;
			int TextX;
			if (mShowUI is mPlanets[0])
			{
				Text = "Take";
				TextX = 18;
			}
			else
			{
				Text = "Buy";
				TextX = 33;
			}
			DrawRect(GuiX + 650 - 50 - 150, GuiY + 95, 150, 30, 0x686868);
			mObelisk.DrawText(Text, GuiX + 650 - 50 - 150 + TextX, GuiY + 95);
			DrawRect(GuiX + 650 - 50 - 150, GuiY + 95 + 35, 150, 30, 0x686868);
			mObelisk.DrawText(Text, GuiX + 650 - 50 - 150 + TextX, GuiY + 95 + 35);
			DrawRect(GuiX + 650 - 50 - 150, GuiY + 95 + 70, 150, 30, 0x686868);
			mObelisk.DrawText(Text, GuiX + 650 - 50 - 150 + TextX, GuiY + 95 + 70);
			if (mTradeResource == -1 )
			{
				if (mShowUI !is mPlanets[0])
				{
					string ResourceName;
					if (mShowUI.GetProduceResource() == 0)
						ResourceName = "Fuel";
					else
						ResourceName = mShowUI.GetProduceResource() == 1 ? "Ammo" : "Ship Parts";
					mObelisk.DrawText("Produces " ~ ResourceName, GuiX + 45, GuiY + 95 + 140);
				}
			}
			else
			{
				mObelisk.DrawText("Sell", GuiX + 45, GuiY + 95 + 135);
				TextX = 0;
				Text = "";
				if (mTradeResource != 0)
				{
					int FuelSell = 10;
					int FuelBuy = cast(int)round(FuelSell * GetExchange(0, mShowUI.GetResource(0),
							mTradeResource, mShowUI.GetResource(mTradeResource)));
					Text = to!string(FuelSell) ~ "~" ~ to!string(FuelBuy);
					if (mCargo[0] != 0)
					{
						DrawRect(GuiX + 650 - 50 - 150, GuiY + 95 + 170, 150, 30, 0x686868);
						mObelisk.DrawText("Trade", GuiX + 650 - 50 - 150 + TextX, GuiY + 95 + 170);
					}
				}
				mObelisk.DrawText("Fuel: " ~ Text, GuiX + 45, GuiY + 95 + 170);

				Text = "";
				if (mTradeResource != 1)
				{
					int AmmoSell = 5;
					int AmmoBuy = cast(int)round(AmmoSell * GetExchange(1, mShowUI.GetResource(1),
							mTradeResource, mShowUI.GetResource(mTradeResource)));
					Text = to!string(AmmoSell) ~ "~" ~ to!string(AmmoBuy);
					if (mCargo[1] != 0)
					{
						DrawRect(GuiX + 650 - 50 - 150, GuiY + 95 + 205, 150, 30, 0x686868);
						mObelisk.DrawText("Trade", GuiX + 650 - 50 - 150 + TextX, GuiY + 95 + 205);
					}
				}
				mObelisk.DrawText("Ammo: " ~ Text, GuiX + 45, GuiY + 95 + 205);

				Text = "";
				if (mTradeResource != 2)
				{
					int ShipPartsSell = 10;
					int ShipPartsBuy = cast(int)round(ShipPartsSell * GetExchange(2, mShowUI.GetResource(2),
							mTradeResource, mShowUI.GetResource(mTradeResource)));
					Text = to!string(ShipPartsSell) ~ "~" ~ to!string(ShipPartsBuy);
					if (mCargo[2] != 0)
					{
						DrawRect(GuiX + 650 - 50 - 150, GuiY + 95 + 240, 150, 30, 0x686868);
						mObelisk.DrawText("Trade", GuiX + 650 - 50 - 150 + TextX, GuiY + 95 + 240);
					}
				}
				mObelisk.DrawText("ShipParts:" ~ Text, GuiX + 45, GuiY + 95 + 240);
			}

			DrawRect(512 - 250 / 2, GuiY + 450 - 20 - 45, 250, 45, 0x686868);
			mObelisk.DrawText("Launch", 512 - 172 / 2, GuiY + 450 - 20 - 45 + 8);

			int CursorX;
			int CursorY;
			mObelisk.GetCursorPos(CursorX, CursorY);
			auto Cursor = mObelisk.GetTexture("Cursor.png");
			auto V0 = new CVector(CursorX - 15, CursorY - 15, 0);
			auto V1 = new CVector(CursorX + 17, CursorY - 15, 0);
			auto V2 = new CVector(CursorX + 17, CursorY + 17, 0);
			auto V3 = new CVector(CursorX - 15, CursorY + 17, 0);
			mObelisk.TriTexture(V0, 0, 0, V1, 1, 0, V2, 1, 1, Cursor);
			mObelisk.TriTexture(V0, 0, 0, V2, 1, 1, V3, 0, 1, Cursor);
		}

		if (mInfoText != "")
		{
			mObelisk.SetAlpha(0.45f);
			DrawRect(0, 0, 1024, 768, 0x000000);
			mObelisk.SetAlpha(1.0f);
			mObelisk.DrawText(mInfoText, 512 - mInfoText.length * 24 / 2, 200);
			if (mGameOver)
				mObelisk.DrawText("Close the window to quit", 194, 500);
			else
				mObelisk.DrawText("Press Enter or Click to continue", 100, 450);
		}
	}



	private float GetExchange(int aResourceSell, int aSellAmount, int aResourceBuy,
			int aBuyAmount)
	{
		assert(aResourceSell != aResourceBuy);
		final switch (aResourceSell)
		{
			case 0:
				return aResourceBuy == 1 ? 1 / FuelAmmoExchange(aSellAmount, aBuyAmount) :
						1 / FuelShipPartsExchange(aSellAmount, aBuyAmount);
			case 1:
				return aResourceBuy == 0 ? FuelAmmoExchange(aBuyAmount, aSellAmount) :
						1 / AmmoShipPartsExchange(aSellAmount, aBuyAmount);
			case 2:
				return aResourceBuy == 0 ? FuelShipPartsExchange(aBuyAmount, aSellAmount) :
						AmmoShipPartsExchange(aBuyAmount, aSellAmount);
		}
	}



	/**
	Calculates the exchange rate between fuel and ammo for the specified amounts of
	fuel and ammo that the planet has.
	*/
	private float FuelAmmoExchange(int aFuel, int aAmmo)
	{
		aFuel = max(aFuel, 10);
		aAmmo = max(aAmmo, 5);
		return (aFuel + 12.5) / aAmmo * (2.0f / 9);
	}



	/**
	Calculates the exchange rate between fuel and ship parts for the specified amounts of
	fuel and ship parts that the planet has.
	*/
	private float FuelShipPartsExchange(int aFuel, int aShipParts)
	{
		aFuel = max(aFuel, 10);
		aShipParts = max(aShipParts, 5);
		return (aFuel + 12.5) / aShipParts * (40.0f / 9);
	}



	/**
	Calculates the exchange rate between ammo and ship parts for the specified amounts of
	ammo and ship parts that the planet has.
	*/
	private float AmmoShipPartsExchange(int aAmmo, int aShipParts)
	{
		aAmmo = max(aAmmo, 5);
		aShipParts = max(aShipParts, 5);
		return (aAmmo + 12.5) / aShipParts * (80.0f / 9);
	}



	private bool ButtonClicked(int aX0, int aY0, int aXSize, int aYSize)
	{
		if (!mObelisk.MouseButtonClicked(false))
			return false;
		int CursorX;
		int CursorY;
		mObelisk.GetCursorPos(CursorX, CursorY);
		return CursorX >= aX0 && CursorX < aX0 + aXSize && CursorY >= aY0 && CursorY < aY0 +
				aYSize;
	}



	private void DrawRect(int aX, int aY, int aXSize, int aYSize, int aColour)
	{
		auto V0 = new CVector(aX, aY, 0);
		auto V1 = new CVector(aX + aXSize, aY, 0);
		auto V2 = new CVector(aX + aXSize, aY + aYSize, 0);
		auto V3 = new CVector(aX, aY + aYSize, 0);
		auto ColourObj = new CColour(((aColour >> 16) & 0xff) / 255.0f,
				((aColour >> 8) & 0xff) / 255.0f, (aColour & 0xff) / 255.0f);
		mObelisk.TriColour(V0, V1, V2, ColourObj);
		mObelisk.TriColour(V0, V2, V3, ColourObj);
	}



	private void GenerateStars()
	{
		for (int i = 0; i < 15000; i++)
		{
			float X = uniform(0.0f, cSpaceSize);
			float Y = uniform(0.0f, cSpaceSize);
			float Intensity = uniform(0.0f, 1.0f) ^^ 2 * 0.4 + 0.5;
			mView.CreateViewObject(new CViewObject(X, Y, 0.9f, 8 * Intensity, 8 * Intensity,
					0, "Star.glsl", Intensity));
		}
	}



	private void GeneratePlanets()
	{
		string[7] Names = [ "Earth", "Xhzhcck", "Zzopop", "Wwwwwwww", "Enola", "Qxdklxdklk",
				"Kghhbghg" ];
		for (int i = 0; i < 7; i++)
		{
			float X;
			float Y;
			while (true)
			{
				X = uniform(0.0f, cSpaceSize);
				Y = uniform(0.0f, cSpaceSize);
				bool Create = true;
				for (int j = 0; j < i; j++)
				{
					if (mView.Distance2(X, Y, mPlanets[j].GetX(), mPlanets[j].GetY()) < 4000 ^^ 2)
					{
						Create = false;
						break;
					}
				}
				if (Create)
					break;
			}
			// 100 = 1 resource / s
			// 0 -> 50 (1 / 2 s)
			// 1 -> 20 (1 / 5 s)
			// 2 -> 10 (1 / 10 s)
			int Resource;
			int Rate;
			final switch (i)
			{
				case 0:
					Resource = -1;
					Rate = 0;
					break;
				case 1:
				case 2:
					Resource = 0;
					Rate = 50;
					break;
				case 3:
				case 4:
					Resource = 1;
					Rate = 20;
					break;
				case 5:
				case 6:
					Resource = 2;
					Rate = 10;
					break;
			}
			mPlanets ~= [ new CPlanet(mView, X, Y, 100, 10, 100, Resource, Rate, Names[i]) ];
		}
	}
}