﻿module CShip;
import std.algorithm, std.math;
import Obelisk.CColour;
import CGameObject, CView, CViewObject;

class CShip : CGameObject
{
	private float mXVel = 0;
	private float mYVel = 0;
	private bool mLanded = true;

	public this(CView aView, float aX, float aY)
	{
		mViewObject = aView.CreateViewObject(new CViewObject(aX, aY, 0, 32, 32, PI * -0.5f,
				"Ship.png"));
	}



	public void Reset(float aX, float aY)
	{
		mViewObject.X = aX;
		mViewObject.Y = aY;
		mViewObject.Angle = PI * -0.5f;
		mXVel = 0;
		mYVel = 0;
		mLanded = true;
	}



	public void Update()
	{
		mViewObject.Move(mXVel / 10, mYVel / 10);
	}



	public void Gravity(float aDeltaX, float aDeltaY)
	{
		float Angle = atan2(-aDeltaY, -aDeltaX);
		float Distance = sqrt(aDeltaX ^^ 2 + aDeltaY ^^ 2);
		float Force = 26 / Distance;
		mXVel += Force * cos(Angle);
		mYVel += Force * sin(Angle);
	}



	public void Thrust()
	{
		float Vel2 = (mXVel ^^ 2 + mYVel ^^ 2) / 10000;
		mXVel = mXVel + cos(mViewObject.Angle) * (1.0f * (1 - Vel2) + 3.0f * Vel2);
		mYVel = mYVel + sin(mViewObject.Angle) * (1.0f * (1 - Vel2) + 3.0f * Vel2);
		Vel2 = mXVel ^^ 2 + mYVel ^^ 2;
		if (Vel2 > 100 ^^ 2)
		{
			float Angle = atan2(mYVel, mXVel);
			mXVel = 100 * cos(Angle);
			mYVel = 100 * sin(Angle);
		}
		mLanded = false;
	}



	public void Turn(int aDirection)
	{
		if (!mLanded)
			mViewObject.Rotate(3 * aDirection * PI / 180);
	}



	public bool IsLanded()
	{
		return mLanded;
	}



	public float GetVelocity()
	{
		return sqrt(mXVel ^^ 2 + mYVel ^^ 2);
	}



	public void Land()
	{
		mXVel = 0;
		mYVel = 0;
		mLanded = true;
	}
}