﻿module CViewObject;
import std.math;
import Obelisk.CColour;
import CView;

enum ERenderMode { Colour, ShaderWithIntensity, Texture }

class CViewObject
{
	public CView Owner;
	public float X;
	public float Y;
	public float Z;
	public float XSize;
	public float YSize;
	public float Angle;
	public ERenderMode RenderMode;
	public CColour Colour;
	public string Texture;
	public string Shader;
	public float Intensity;
	public CColour MinimapColour = null;

	public this(float aX, float aY, float aZ, float aXSize, float aYSize, float aAngle, CColour aColour)
	{
		X = aX;
		Y = aY;
		Z = aZ;
		XSize = aXSize;
		YSize = aYSize;
		Angle = aAngle;
		RenderMode = ERenderMode.Colour;
		Colour = aColour;
	}

	public this(float aX, float aY, float aZ, float aXSize, float aYSize, float aAngle, string aTexture)
	{
		X = aX;
		Y = aY;
		Z = aZ;
		XSize = aXSize;
		YSize = aYSize;
		Angle = aAngle;
		RenderMode = ERenderMode.Texture;
		Texture = aTexture;
	}

	public this(float aX, float aY, float aZ, float aXSize, float aYSize, float aAngle,
			string aShader, float aIntensity)
	{
		X = aX;
		Y = aY;
		Z = aZ;
		XSize = aXSize;
		YSize = aYSize;
		Angle = aAngle;
		RenderMode = ERenderMode.ShaderWithIntensity;
		Shader = aShader;
		Intensity = aIntensity;
	}



	public void Move(float aXVel, float aYVel)
	{
		X += aXVel;
		if (X < 0)
			X += Owner.GetViewSize();
		else if (X >= Owner.GetViewSize())
			X -= Owner.GetViewSize();
		Y += aYVel;
		if (Y < 0)
			Y += Owner.GetViewSize();
		else if (Y >= Owner.GetViewSize())
			Y -= Owner.GetViewSize();
	}



	public void Rotate(float aAngle)
	{
		Angle += aAngle;
		if (Angle <= -PI)
			Angle += PI * 2;
		else if (Angle > PI)
			Angle -= PI * 2;
	}
}